// Generated by gir (https://github.com/gtk-rs/gir @ 73df75128d55)
// from ../gir-files (@ 8be196040235)
// DO NOT EDIT

#include "manual.h"
#include <stdalign.h>
#include <stdio.h>

int main() {
    printf("%s;%zu;%zu\n", "JSCCheckSyntaxMode", sizeof(JSCCheckSyntaxMode), alignof(JSCCheckSyntaxMode));
    printf("%s;%zu;%zu\n", "JSCCheckSyntaxResult", sizeof(JSCCheckSyntaxResult), alignof(JSCCheckSyntaxResult));
    printf("%s;%zu;%zu\n", "JSCClassVTable", sizeof(JSCClassVTable), alignof(JSCClassVTable));
    printf("%s;%zu;%zu\n", "JSCContext", sizeof(JSCContext), alignof(JSCContext));
    printf("%s;%zu;%zu\n", "JSCContextClass", sizeof(JSCContextClass), alignof(JSCContextClass));
    printf("%s;%zu;%zu\n", "JSCException", sizeof(JSCException), alignof(JSCException));
    printf("%s;%zu;%zu\n", "JSCExceptionClass", sizeof(JSCExceptionClass), alignof(JSCExceptionClass));
    printf("%s;%zu;%zu\n", "JSCOptionType", sizeof(JSCOptionType), alignof(JSCOptionType));
    printf("%s;%zu;%zu\n", "JSCTypedArrayType", sizeof(JSCTypedArrayType), alignof(JSCTypedArrayType));
    printf("%s;%zu;%zu\n", "JSCValue", sizeof(JSCValue), alignof(JSCValue));
    printf("%s;%zu;%zu\n", "JSCValueClass", sizeof(JSCValueClass), alignof(JSCValueClass));
    printf("%s;%zu;%zu\n", "JSCValuePropertyFlags", sizeof(JSCValuePropertyFlags), alignof(JSCValuePropertyFlags));
    printf("%s;%zu;%zu\n", "JSCVirtualMachine", sizeof(JSCVirtualMachine), alignof(JSCVirtualMachine));
    printf("%s;%zu;%zu\n", "JSCVirtualMachineClass", sizeof(JSCVirtualMachineClass), alignof(JSCVirtualMachineClass));
    printf("%s;%zu;%zu\n", "JSCWeakValue", sizeof(JSCWeakValue), alignof(JSCWeakValue));
    printf("%s;%zu;%zu\n", "JSCWeakValueClass", sizeof(JSCWeakValueClass), alignof(JSCWeakValueClass));
    return 0;
}
