// This file was generated by gir (https://github.com/gtk-rs/gir)
// from
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::{AuthDomain, ServerMessage};
use glib::{
    prelude::*,
    signal::{connect_raw, SignalHandlerId},
    translate::*,
};
use std::{boxed::Box as Box_, fmt, mem::transmute};

glib::wrapper! {
    #[doc(alias = "SoupAuthDomainBasic")]
    pub struct AuthDomainBasic(Object<ffi::SoupAuthDomainBasic, ffi::SoupAuthDomainBasicClass>) @extends AuthDomain;

    match fn {
        type_ => || ffi::soup_auth_domain_basic_get_type(),
    }
}

impl AuthDomainBasic {
    //#[doc(alias = "soup_auth_domain_basic_new")]
    //pub fn new(optname1: &str, : /*Unknown conversion*//*Unimplemented*/Basic: VarArgs) -> AuthDomainBasic {
    //    unsafe { TODO: call ffi:soup_auth_domain_basic_new() }
    //}

    #[doc(alias = "soup_auth_domain_basic_set_auth_callback")]
    pub fn set_auth_callback<
        P: Fn(&AuthDomainBasic, &ServerMessage, &str, &str) -> bool + 'static,
    >(
        &self,
        callback: P,
    ) {
        let callback_data: Box_<P> = Box_::new(callback);
        unsafe extern "C" fn callback_func<
            P: Fn(&AuthDomainBasic, &ServerMessage, &str, &str) -> bool + 'static,
        >(
            domain: *mut ffi::SoupAuthDomainBasic,
            msg: *mut ffi::SoupServerMessage,
            username: *const libc::c_char,
            password: *const libc::c_char,
            user_data: glib::ffi::gpointer,
        ) -> glib::ffi::gboolean {
            let domain = from_glib_borrow(domain);
            let msg = from_glib_borrow(msg);
            let username: Borrowed<glib::GString> = from_glib_borrow(username);
            let password: Borrowed<glib::GString> = from_glib_borrow(password);
            let callback: &P = &*(user_data as *mut _);
            (*callback)(&domain, &msg, username.as_str(), password.as_str()).into_glib()
        }
        let callback = Some(callback_func::<P> as _);
        unsafe extern "C" fn dnotify_func<
            P: Fn(&AuthDomainBasic, &ServerMessage, &str, &str) -> bool + 'static,
        >(
            data: glib::ffi::gpointer,
        ) {
            let _callback: Box_<P> = Box_::from_raw(data as *mut _);
        }
        let destroy_call3 = Some(dnotify_func::<P> as _);
        let super_callback0: Box_<P> = callback_data;
        unsafe {
            ffi::soup_auth_domain_basic_set_auth_callback(
                self.to_glib_none().0,
                callback,
                Box_::into_raw(super_callback0) as *mut _,
                destroy_call3,
            );
        }
    }

    //#[doc(alias = "auth-data")]
    //pub fn auth_data(&self) -> /*Unimplemented*/Basic: Pointer {
    //    ObjectExt::property(self, "auth-data")
    //}

    //#[doc(alias = "auth-data")]
    //pub fn set_auth_data(&self, auth_data: /*Unimplemented*/Basic: Pointer) {
    //    ObjectExt::set_property(self,"auth-data", auth_data)
    //}

    #[doc(alias = "auth-data")]
    pub fn connect_auth_data_notify<F: Fn(&Self) + 'static>(&self, f: F) -> SignalHandlerId {
        unsafe extern "C" fn notify_auth_data_trampoline<F: Fn(&AuthDomainBasic) + 'static>(
            this: *mut ffi::SoupAuthDomainBasic,
            _param_spec: glib::ffi::gpointer,
            f: glib::ffi::gpointer,
        ) {
            let f: &F = &*(f as *const F);
            f(&from_glib_borrow(this))
        }
        unsafe {
            let f: Box_<F> = Box_::new(f);
            connect_raw(
                self.as_ptr() as *mut _,
                b"notify::auth-data\0".as_ptr() as *const _,
                Some(transmute::<_, unsafe extern "C" fn()>(
                    notify_auth_data_trampoline::<F> as *const (),
                )),
                Box_::into_raw(f),
            )
        }
    }
}

impl fmt::Display for AuthDomainBasic {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.write_str("AuthDomainBasic")
    }
}
