pub mod asin_const;
pub mod asinh_const;
pub mod atan_const;
pub mod cbrt_const;
pub mod exp_const;
pub mod fact_const;
pub mod ln_const;
pub mod sin_const;
pub mod sqrt_const;
pub mod tan_const;
pub mod tanh_const;
