// This file was generated by gir (https://github.com/gtk-rs/gir)
// from
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::{HSTSEnforcer, SessionFeature};
use glib::{prelude::*, translate::*};
use std::fmt;

glib::wrapper! {
    #[doc(alias = "SoupHSTSEnforcerDB")]
    pub struct HSTSEnforcerDB(Object<ffi::SoupHSTSEnforcerDB, ffi::SoupHSTSEnforcerDBClass>) @extends HSTSEnforcer, @implements SessionFeature;

    match fn {
        type_ => || ffi::soup_hsts_enforcer_db_get_type(),
    }
}

impl HSTSEnforcerDB {
    #[doc(alias = "soup_hsts_enforcer_db_new")]
    pub fn new(filename: &str) -> HSTSEnforcerDB {
        assert_initialized_main_thread!();
        unsafe {
            HSTSEnforcer::from_glib_full(ffi::soup_hsts_enforcer_db_new(filename.to_glib_none().0))
                .unsafe_cast()
        }
    }

    pub fn filename(&self) -> Option<glib::GString> {
        ObjectExt::property(self, "filename")
    }
}

impl fmt::Display for HSTSEnforcerDB {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.write_str("HSTSEnforcerDB")
    }
}
